

import os
import json
from functools import partial
from typing import Optional

from .httpapi import MiraiHttpApi
from .message.chain import MessageChain
from .message.messages import Message

class MiraiSession(object):
    def __init__(self, api: MiraiHttpApi, account: str, authKey: str):
        self.llapi = api
        self.authKey = authKey
        self.account = account
        self.sessionKey = None

    def auth(self):
        sess_response = self.llapi.auth(authKey=self.authKey)
        self.sessionKey = sess_response['session']
        verify_response = self.llapi.verify(
            sessionKey=self.sessionKey, qq=self.account)
        if verify_response['msg'] != "success":
            raise Exception(verify_response['msg'])

    def leave(self):
        release_response = self.llapi.release(
            sessionKey=self.sessionKey, qq=self.account)
        self.sessionKey = None

        if release_response['msg'] != "success":
            raise Exception(release_response['msg'])

    def __enter__(self):
        self.auth()
        # print(f'打开session: {self.sessionKey}')
        return self

    def __exit__(self, type, value, trace):
        # print(f'关闭session: {self.sessionKey}')
        self.leave()
    
    def recall(self, target:int):
        '''
        sessionKey	String	已经激活的Session
        target	    Int	需要撤回的消息的messageId
        '''
        return self.llapi.recall(sessionKey=self.sessionKey, target=target)

    def sendGroupMessage(self, target:int, messageChain: MessageChain, quote:Optional[int]=None):
        '''
        sessionKey	    String	已经激活的Session
        target	        Long	可选，发送消息目标群的群号
        group	        Long	可选，target与group中需要有一个参数不为空，当target不为空时group将被忽略，同target
        quote	        Int	    引用一条消息的messageId进行回复
        messageChain	Array	消息链，是一个消息对象构成的数组
        '''
        if quote is None:
            ret = self.llapi.sendGroupMessage(
                sessionKey=self.sessionKey,
                target=target,
                messageChain=messageChain.dict()
            )
        else:
            ret = self.llapi.sendGroupMessage(
                sessionKey=self.sessionKey,
                target=target,
                quote=quote,
                messageChain=messageChain.dict()
            )
        
        return Message(ret['messageId'], messageChain)

    def sendFriendMessage(self, target:int, messageChain: MessageChain, quote:Optional[int]=None):
        '''
        sessionKey	    String	已经激活的Session
        target	        Long	可选，发送消息目标群的群号
        qq  	        Long	可选，target与qq中需要有一个参数不为空，当target不为空时qq将被忽略，同target
        quote	        Int	    引用一条消息的messageId进行回复
        messageChain	Array	消息链，是一个消息对象构成的数组
        '''
        if quote is None:
            ret = self.llapi.sendFriendMessage(
                sessionKey=self.sessionKey,
                target=target,
                messageChain=messageChain.dict()
            )
        else:
            ret = self.llapi.sendFriendMessage(
                sessionKey=self.sessionKey,
                target=target,
                quote=quote,
                messageChain=messageChain.dict()
            )
        
        return Message(ret['messageId'], messageChain)
        
    def peekLatestMessage(self, count:int):
        '''
        sessionKey	你的session key
        count   	获取消息和事件的数量
        '''
        ret = self.llapi.peekLatestMessage(sessionKey=self.sessionKey, count=count)
        data = ret['data']
        return data

    def friendList(self):
        '''
        使用此方法获取bot的好友列表
        '''
        ret = self.llapi.friendList(sessionKey=self.sessionKey)
        return ret

    def groupList(self):
        '''
        使用此方法获取bot的群列表
        '''
        ret = self.llapi.groupList(sessionKey=self.sessionKey)
        return ret

    def memberList(self, target:int):
        '''
        使用此方法获取bot的群列表
        '''
        ret = self.llapi.memberList(sessionKey=self.sessionKey, target=target)
        return ret



from typing import (
    Any,
    Dict,
    Iterable,
    List,
    NoReturn,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    Optional
)
import copy
from .element import Element, InternalElement, ExternalElement

from pydantic import BaseModel

class MessageChain(BaseModel):
    messages: Sequence[Element]

    @classmethod
    def create(cls, elements: Sequence[Element]) -> "MessageChain":
        """从传入的序列(可以是元组 tuple, 也可以是列表 list) 创建消息链.

        Args:
            elements (Sequence[T]): 包含且仅包含消息元素的序列

        Returns:
            MessageChain: 以传入的序列作为所承载消息的消息链
        """
        if not isinstance(elements, Sequence):
            raise Exception("参数必须是序列类型")
        return cls(messages=elements)

    @classmethod
    def parse_obj(cls: Type["MessageChain"], obj: List[Element]) -> "MessageChain":
        """内部接口, 会自动将作为外部态的消息元素转为内部态.

        Args:
            obj (List[T]): 需要反序列化的对象

        Returns:
            MessageChain: 内部承载有尽量有效的内部态消息元素的消息链
        """
        handled_elements = []
        for i in obj:
            if isinstance(i, InternalElement):
                handled_elements.append(i)
            elif isinstance(i, ExternalElement):
                for ii in InternalElement.__subclasses__():
                    if ii.__name__ == i.__class__.__name__:
                        handled_elements.append(ii.fromExternal(i))
            elif isinstance(i, dict) and "type" in i:
                for ii in ExternalElement.__subclasses__():
                    if ii.__name__ == i["type"]:
                        for iii in InternalElement.__subclasses__():
                            if iii.__name__ == i["type"]:
                                handled_elements.append(
                                    iii.fromExternal(ii.parse_obj(i))
                                )
        return cls(messages=tuple(handled_elements))  # 默认是不可变型

    def dict(self):
        return [m.dict() for m in self.messages]

    
    def subchain(self, item: slice, ignore_text_index: bool = False) -> "MessageChain":
        """对消息链执行分片操作

        Args:
            item (slice): 这个分片的 `start` 和 `end` 的 Type Annotation 都是 `Optional[MessageIndex]`

        Raises:
            TypeError: TextIndex 取到了错误的位置

        Returns:
            MessageChain: 分片后得到的新消息链, 绝对是原消息链的子集.
        """
        from .element import Plain

        result = copy.copy(self.messages)
        if item.start:
            first_slice = result[item.start[0] :]
            if item.start[1] is not None and first_slice:  # text slice
                if not isinstance(first_slice[0], Plain):
                    if not ignore_text_index:
                        raise TypeError(
                            "the sliced chain does not starts with a Plain: {}".format(
                                first_slice[0]
                            )
                        )
                    else:
                        result = first_slice
                else:
                    final_text = first_slice[0].text[item.start[1] :]
                    result = [
                        *([Plain(final_text)] if final_text else []),
                        *first_slice[1:],
                    ]
            else:
                result = first_slice
        if item.stop:
            first_slice = result[: item.stop[0]]
            if item.stop[1] is not None and first_slice:  # text slice
                if not isinstance(first_slice[-1], Plain):
                    raise TypeError(
                        "the sliced chain does not ends with a Plain: {}".format(
                            first_slice[-1]
                        )
                    )
                final_text = first_slice[-1].text[: item.stop[1]]
                result = [
                    *first_slice[:-1],
                    *([Plain(final_text)] if final_text else []),
                ]
            else:
                result = first_slice
        return MessageChain.create(result)
    
    def has(self, element_class: Element) -> bool:
        """判断消息链中是否含有特定类型的消息元素

        Args:
            element_class (T): 需要判断的消息元素的类型, 例如 "Plain", "At", "Image" 等.

        Returns:
            bool: 判断结果
        """
        return element_class in [type(i) for i in self.messages]

    def get(self, element_class: Element) -> List[Element]:
        """获取消息链中所有特定类型的消息元素

        Args:
            element_class (T): 指定的消息元素的类型, 例如 "Plain", "At", "Image" 等.

        Returns:
            List[T]: 获取到的符合要求的所有消息元素; 另: 可能是空列表([]).
        """
        return [i for i in self.messages if type(i) is element_class]

    def getOne(self, element_class: Element, index: int) -> Element:
        """获取消息链中第 index + 1 个特定类型的消息元素

        Args:
            element_class (Type[Element]): 指定的消息元素的类型, 例如 "Plain", "At", "Image" 等.
            index (int): 索引, 从 0 开始数

        Returns:
            T: 消息链第 index + 1 个特定类型的消息元素
        """
        return self.get(element_class)[index]

    def getFirst(self, element_class: Element) -> Element:
        """获取消息链中第 1 个特定类型的消息元素

        Args:
            element_class (Type[Element]): 指定的消息元素的类型, 例如 "Plain", "At", "Image" 等.

        Returns:
            T: 消息链第 1 个特定类型的消息元素
        """
        return self.getOne(element_class, 0)

    def asDisplay(self) -> str:
        """获取以字符串形式表示的消息链, 且趋于通常你见到的样子.

        Returns:
            str: 以字符串形式表示的消息链
        """
        return "".join(i.asDisplay() for i in self.messages)
    
    def __getitem__(self, item: Union[Type[Element], slice]):
        if isinstance(item, slice):
            return self.subchain(item)
        elif issubclass(item, Element):
            return self.get(item)
        else:
            raise NotImplementedError(
                "{0} is not allowed for item getting".format(type(item))
            )



import json
from functools import partial
from typing import Optional

import requests
from urllib.parse import urljoin
from urllib.parse import quote
from tenacity import retry


class MiraiHttpApi(object):
    '''
    底层HTTP接口
    文档地址：https://github.com/project-mirai/mirai-api-http/blob/master/docs/API.md
    '''

    def __init__(self, host: str):
        self.host = host

    def check_code(self, code: int):
        if not isinstance(code, int):
            code = int(code)

        if code == 0:
            return
        elif code == 1:
            raise Exception("错误的auth key")
        elif code == 2:
            raise Exception("指定的Bot不存在")
        elif code == 3:
            raise Exception("Session失效或不存在")
        elif code == 4:
            raise Exception("Session未认证(未激活)")
        elif code == 5:
            raise Exception("发送消息目标不存在(指定对象不存在)")
        elif code == 6:
            raise Exception("指定文件不存在，出现于发送本地图片")
        elif code == 10:
            raise Exception("无操作权限，指Bot没有对应操作的限权")
        elif code == 20:
            raise Exception("Bot被禁言，指Bot当前无法向指定群发送消息")
        elif code == 30:
            raise Exception("消息过长")
        elif code == 400:
            raise Exception("错误的访问，如参数错误等")
        return

    # low level apis
    def query_get(self, api_name: str, **kwargs):
        url = urljoin(self.host, '/' + api_name)
        params = [k + "=" + quote(str(v)) for k, v in kwargs.items()]
        if len(params) > 0:
            url = url + '?' + '&'.join(params)
        response = requests.get(url)
        if response.status_code != 200:
            raise Exception("异常状态码" + str(response.status_code))
        obj = json.loads(response.text)
        if "code" in obj:
            self.check_code(obj["code"])

        return obj

    def query_post(self, api_name: str, **kwargs):
        req_params = kwargs
        url = urljoin(self.host, '/' + api_name)
        response = requests.post(url, json=req_params)
        if response.status_code != 200:
            raise Exception("异常状态码" + str(response.status_code))
        obj = json.loads(response.text)
        if "code" in obj:
            self.check_code(obj["code"])

        return obj

    @retry
    def __getattr__(self, name: str):
        get_list = ['about', 'fetchMessage', 'fetchLatestMessage', 'peekMessage',
                    'peekLatestMessage', 'friendList', 'groupList', 'memberList']
        if name in get_list:
            return partial(self.query_get, name)
        else:
            return partial(self.query_post, name)



import json
from typing import Callable, NoReturn, Dict, Union, Type, List
import time, threading
import concurrent
from collections import defaultdict

from mirai.api import MiraiApi
from mirai.message.chain import MessageChain
from mirai.message.messages import Message
from mirai.event import BaseEvent, GroupMessageEvent, FriendMessageEvent
from mirai.sender import BaseSender, Friend, Group, GroupMessageSender

class Listener(object):
    def __init__(self, event: str, handler:Callable):
        self.event = event
        self.handler = handler

class MiraiApp(object):
    def __init__(self, host:str, authKey:str, account:str, logger=None):
        self.host = host
        self.authKey = authKey
        self.account = account
        self.api = MiraiApi(host, authKey)
        self.session = None
        self.listeners = defaultdict(list)

    @staticmethod
    def dispatch_event(app, event: Dict):
        if event['type'] not in ['GroupMessage', 'FriendMessage']:
            print(event)
        event_type = event['type']
        for listener in app.listeners[event_type]:
            for subEventClass in BaseEvent.__subclasses__():
                subEventClassName = event_type + 'Event' if not event_type.endswith('Event') else event_type
                if subEventClass.__name__ == subEventClassName:
                    handler_event = subEventClass.parse_obj(event)
                    break
            listener.handler(app, handler_event)
    
    def fn_message_thread(self):
        with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
            history = []
            quick_history_set = set()
            loop_count = 0
            while True:
                if loop_count % 1000 == 0:
                    quick_history_set = set([json.dumps(d) for d in history[-50:]])

                ret = self.session.peekLatestMessage(count=50)
                for each_event in ret:
                    event_id = json.dumps(each_event)
                    if event_id not in quick_history_set:
                        history.append(each_event)
                        quick_history_set.add(event_id)
                        executor.submit(self.dispatch_event, self, each_event)

                time.sleep(1)
                loop_count += 1

    def blocking_start(self) -> NoReturn:
        self.fn_message_thread()

    def auth(self) -> NoReturn:
        self.session = self.api.get_session(self.account)
        self.session.auth()

    def leave(self) -> NoReturn:
        self.session.leave()
        self.session = None

    def __enter__(self):
        self.auth()
        return self

    def __exit__(self, type, value, trace) -> NoReturn:
        self.leave()

    def register(self, event: Union[str, Type], handler: Callable) -> NoReturn:
        if isinstance(event, str):
            self.listeners[event].append(Listener(event, handler))
        elif issubclass(event, BaseEvent):
            self.listeners[event.name].append(Listener(event.name, handler))
        else:
            raise Exception("parameter event must be str or BaseEvent")

    def sendGroupMessage(self, target: int, message: MessageChain) -> Message:
        return self.session.sendGroupMessage(target=target, messageChain=message)
    
    def sendFriendMessage(self, target: int, message: MessageChain) -> Message:
        return self.session.sendFriendMessage(target=target, messageChain=message)

    def friendList(self) -> List[Friend]:
        ret = self.session.friendList()
        return [Friend.parse_obj(f) for f in ret]




from typing import Dict

from mirai.message.chain import MessageChain
from mirai.message.messages import Message
from mirai.sender import BaseSender, Friend, GroupMessageSender


class BaseEvent(object):
    pass


class MemberCardChangeEvent(BaseEvent):
    name = 'MemberCardChangeEvent'

    def __init__(self, origin: str, new: str, current: str, member: GroupMessageSender, operator: None):
        self.origin = origin
        self.new = new
        self.current = current
        self.member = member
        self.operator = operator


class BotMuteEvent(BaseEvent):
    name = 'BotMuteEvent'

    def __init__(self, durationSeconds: int, operator: GroupMessageSender):
        self.durationSeconds = durationSeconds
        self.operator = operator

    @classmethod
    def parse_obj(cls, obj: Dict) -> "BotMuteEvent":
        return BotMuteEvent(
            durationSeconds=obj['durationSeconds'],
            operator=GroupMessageSender.parse_obj(obj['operator']))


class BotUnmuteEvent(BaseEvent):
    name = 'BotUnmuteEvent'

    def __init__(self, operator: GroupMessageSender):
        self.operator = operator

    @classmethod
    def parse_obj(cls, obj: Dict) -> "BotUnmuteEvent":
        return BotUnmuteEvent(
            operator=GroupMessageSender.parse_obj(obj['operator']))


class BotReloginEvent(BaseEvent):
    name = 'BotReloginEvent'

    def __init__(self, qq: int):
        self.qq = qq

    @classmethod
    def parse_obj(cls, obj: Dict) -> "BotReloginEvent":
        return BotReloginEvent(qq=obj['qq'])


class BotOnlineEvent(BaseEvent):
    name = 'BotOnlineEvent'

    def __init__(self, qq: int):
        self.qq = qq

    @classmethod
    def parse_obj(cls, obj: Dict) -> "BotOnlineEvent":
        return BotOnlineEvent(qq=obj['qq'])


class BotOfflineEventDropped(BaseEvent):
    name = 'BotOfflineEventDropped'

    def __init__(self, qq: int):
        self.qq = qq

    @classmethod
    def parse_obj(cls, obj: Dict) -> "BotOfflineEventDropped":
        return BotOfflineEventDropped(qq=obj['qq'])


class GroupMessageEvent(BaseEvent):
    name = 'GroupMessage'

    def __init__(self, messageChain: MessageChain, sender: GroupMessageSender):
        self.messageChain = messageChain
        self.sender = sender

    @classmethod
    def parse_obj(cls, obj: Dict) -> "GroupMessageEvent":
        return GroupMessageEvent(
            messageChain=MessageChain.parse_obj(obj['messageChain']),
            sender=GroupMessageSender.parse_obj(obj['sender']))


class FriendMessageEvent(BaseEvent):
    name = 'FriendMessage'

    def __init__(self, messageChain: MessageChain, sender: Friend):
        self.messageChain = messageChain
        self.sender = sender

    @classmethod
    def parse_obj(cls, obj: Dict) -> "FriendMessageEvent":
        return FriendMessageEvent(
            messageChain=MessageChain.parse_obj(obj['messageChain']),
            sender=Friend.parse_obj(obj['sender']))



from typing import NoReturn, Any, Dict


class BaseSender(object):
    pass


class Group(object):
    def __init__(self, id: int, name: str, permission: str):
        self.id = id
        self.name = name
        self.permission = permission


class GroupMessageSender(object):
    def __init__(self: 'GroupMessageSender', id: int, memberName: str, permission: str, group: Group) -> NoReturn:
        self.id = id
        self.memberName = memberName
        self.permission = permission
        self.group = group

    @classmethod
    def parse_obj(cls, obj: Dict) -> "GroupMessageSender":
        return GroupMessageSender(
            id=obj['id'],
            memberName=obj['memberName'],
            permission=obj['permission'],
            group=Group(**obj['group'])
        )


class Friend(object):
    def __init__(self: 'Friend', id: int, nickname: str, remark: str) -> NoReturn:
        self.id = id
        self.nickname = nickname
        self.remark = remark

    @classmethod
    def parse_obj(cls, obj: Dict) -> "Friend":
        return Friend(
            id=obj['id'],
            nickname=obj['nickname'],
            remark=obj['remark'])




import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mirai-py", # Replace with your own username
    version="0.1a1.dev1",
    author="Jun Wang",
    author_email="jstzwj@aliyun.com",
    description="Mirai python API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lemon-chat/mirai-py",
    project_urls={
        "Bug Tracker": "https://github.com/lemon-chat/mirai-py/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    packages=setuptools.find_packages(),
    python_requires='>=3.7'
)


from enum import Enum
from typing import NoReturn, Optional, Union
import datetime

from pydantic import BaseModel

class Element(BaseModel):
    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))


class InternalElement(Element):
    def toExternal(self) -> "ExternalElement":
        """可以为异步方法"""
        pass
    
    @staticmethod
    def fromExternal(cls, external_element) -> "InternalElement":
        """可以为异步方法"""
        pass

    def asDisplay(self) -> str:
        return ""

    def asSerializationString(self) -> str:
        return ""


class ExternalElement(Element):
    pass


class ShadowElement(Element):
    pass

class Plain(InternalElement, ExternalElement):
    type: str = "Plain"
    text: str

    def __init__(self, text: str, *_, **__) -> NoReturn:
        """实例化一个 Plain 消息元素, 用于承载消息中的文字.

        Args:
            text (str): 元素所包含的文字
        """
        super().__init__(text=text)

    def toExternal(self) -> "Plain":
        return self

    @classmethod
    def fromExternal(_, external_element) -> "Plain":
        return external_element

    def asDisplay(self) -> str:
        return self.text

    def asSerializationString(self) -> str:
        return self.text

class Source(InternalElement, ExternalElement):
    "表示消息在一个特定聊天区域内的唯一标识"
    id: int
    time: datetime.datetime

    @classmethod
    def fromExternal(_, external_element) -> "Source":
        return external_element

    def asSerializationString(self) -> str:
        return f"[mirai:source:{self.id},{int(self.time.timestamp())}]"

    class Config:
        json_encoders = {
            datetime: lambda v: int(v.timestamp()),
        }

'''
class Quote(InternalElement, ExternalElement):
    "表示消息中回复其他消息/用户的部分, 通常包含一个完整的消息链(`origin` 属性)"
    id: int
    groupId: int
    senderId: int
    targetId: int
    origin: "MessageChain"

    @validator("origin", pre=True, allow_reuse=True)
    def _(cls, v):
        from ..chain import MessageChain

        return MessageChain.parse_obj(v)

    @classmethod
    def fromExternal(_, external_element) -> "Quote":
        return external_element

    def asSerializationString(self) -> str:
        return f" [mirai:quote:{self.id}]"


class At(InternalElement, ExternalElement):
    """该消息元素用于承载消息中用于提醒/呼唤特定用户的部分."""

    type: str = "At"
    target: int
    display: Optional[str] = None

    def __init__(self, target: int, **kwargs) -> None:
        """实例化一个 At 消息元素, 用于承载消息中用于提醒/呼唤特定用户的部分.

        Args:
            target (int): 需要提醒/呼唤的特定用户的 QQ 号(或者说 id.)
        """
        super().__init__(target=target, **kwargs)

    def asDisplay(self) -> str:
        return str(self.display) if self.display else f"@{self.target}"

    def toExternal(self) -> "At":
        try:
            if image_method.get() != UploadMethods.Group:
                raise InvaildArgument(
                    "you cannot use this element in this method: {0}".format(
                        image_method.get().value
                    )
                )
        except LookupError:
            pass
        return self

    def asSerializationString(self) -> str:
        return f"[mirai:at:{self.target},{self.display}]"

    @classmethod
    def fromExternal(cls, external_element) -> "At":
        return external_element


class AtAll(InternalElement, ExternalElement):
    "该消息元素用于群组中的管理员提醒群组中的所有成员"
    type: str = "AtAll"

    def asDisplay(self) -> str:
        return "@全体成员"

    def toExternal(self) -> "AtAll":
        try:
            if image_method.get() != UploadMethods.Group:
                raise InvaildArgument(
                    "you cannot use this element in this method: {0}".format(
                        image_method.get().value
                    )
                )
        except LookupError:
            pass
        return self

    @classmethod
    def fromExternal(cls, external_element) -> "AtAll":
        return external_element

    def asSerializationString(self) -> str:
        return "[mirai:atall]"


class Face(InternalElement, ExternalElement):
    "表示消息中所附带的表情, 这些表情大多都是聊天工具内置的."
    type: str = "Face"
    faceId: int
    name: Optional[str] = None

    def toExternal(self) -> "Face":
        return self

    @classmethod
    def fromExternal(cls, external_element) -> "Face":
        return external_element

    def asDisplay(self) -> str:
        return "[表情]"

    def asSerializationString(self) -> str:
        return f"[mirai:face:{self.faceId}]"


class ImageType(Enum):
    Friend = "Friend"
    Group = "Group"
    Temp = "Temp"
    Unknown = "Unknown"


image_upload_method_type_map = {
    UploadMethods.Friend: ImageType.Friend,
    UploadMethods.Group: ImageType.Group,
    UploadMethods.Temp: ImageType.Temp,
}


class ShadowImage(InternalElement, ExternalElement, ShadowElement):
    method: Optional[UploadMethods]
    is_flash: bool = False

    def asFlash(self):
        self.is_flash = True
        return self

    def fromExternal(cls, external_element) -> "InternalElement":
        return external_element

    class Config:
        allow_mutation = True


class Image_LocalFile(ShadowImage):
    filepath: Path

    def __init__(self, filepath: Path, method: Optional[UploadMethods] = None) -> None:
        super().__init__(filepath=filepath, method=method)

    async def toExternal(self):
        app = application.get()
        try:
            methodd = self.method or image_method.get()
        except LookupError:
            raise ValueError(
                "you should give the 'method' for upload when you are out of the event receiver."
            )
        if not self.is_flash:
            return await app.uploadImage(
                self.filepath.read_bytes(), methodd, return_external=True
            )
        else:
            return FlashImage.fromExternal(
                await app.uploadImage(
                    self.filepath.read_bytes(), methodd, return_external=True
                )
            ).toExternal()

    async def getReal(self, method: UploadMethods) -> "Image":
        """从本 Shadow Element 中生成一真正的 Image 对象.
        Args:
            method (UploadMethods): 所需求的图片的上传类型, 具体请阅读 UploadMethods 的相关文档.
        Raises:
            ClientResponseError: HTTP 网络请求错误
        Returns:
            Image: 所生成的, 真正的 Image 对象.
        """
        app = application.get()
        return await app.uploadImage(
            self.filepath.read_bytes(), method, return_external=True
        )


class Image_UnsafeBytes(ShadowImage):
    image_bytes: bytes

    def __init__(
        self, image_bytes: bytes, method: Optional[UploadMethods] = None
    ) -> None:
        super().__init__(image_bytes=image_bytes, method=method)

    async def toExternal(self):
        app = application.get()
        try:
            methodd = self.method or image_method.get()
        except LookupError:
            raise ValueError(
                "you should give the 'method' for upload when you are out of the event receiver."
            )
        if not self.is_flash:
            return await app.uploadImage(
                self.image_bytes, methodd, return_external=True
            )
        else:
            return FlashImage.fromExternal(
                await app.uploadImage(self.image_bytes, methodd, return_external=True)
            ).toExternal()

    async def getReal(self, method: UploadMethods) -> "Image":
        """从本 Shadow Element 中生成一真正的 Image 对象.
        Args:
            method (UploadMethods): 所需求的图片的上传类型, 具体请阅读 UploadMethods 的相关文档.
        Raises:
            ClientResponseError: HTTP 网络请求错误
        Returns:
            Image: 所生成的, 真正的 Image 对象.
        """
        app = application.get()
        return await app.uploadImage(self.image_bytes, method)


class Image_NetworkAddress(ShadowImage):
    url: str

    def __init__(self, url: str, method: Optional[UploadMethods] = None) -> None:
        super().__init__(url=url, method=method)

    async def toExternal(self):
        app = application.get()
        try:
            methodd = self.method or image_method.get()
        except LookupError:
            raise ValueError(
                "you should give the 'method' for upload when you are out of the event receiver."
            )

        async with app.session.get(self.url) as response:
            response.raise_for_status()
            if not self.is_flash:
                return await app.uploadImage(
                    await response.read(), methodd, return_external=True
                )
            else:
                return FlashImage.fromExternal(
                    await app.uploadImage(
                        await response.read(), methodd, return_external=True
                    )
                ).toExternal()

    async def getReal(self, method: UploadMethods) -> "Image":
        """从本 Shadow Element 中生成一真正的 Image 对象.
        Args:
            method (UploadMethods): 所需求的图片的上传类型, 具体请阅读 UploadMethods 的相关文档.
        Raises:
            ClientResponseError: HTTP 网络请求错误
        Returns:
            Image: 所生成的, 真正的 Image 对象.
        """
        app = application.get()
        async with aiohttp.ClientSession() as session:
            async with session.get(self.url) as response:
                response.raise_for_status()
                return await app.uploadImage(await response.read(), method)


class Image(InternalElement):
    "该消息元素用于承载消息中所附带的图片."
    imageId: Optional[str] = None
    url: Optional[str] = None
    path: Optional[str] = None
    type: Optional[ImageType]

    @validator("type", always=True, allow_reuse=True)
    def _(cls, v, values) -> ImageType:
        if v:
            return v
        if "imageId" not in values:
            return ImageType.Unknown
        if not values["imageId"]:
            return ImageType.Unknown
        if values["imageId"].startswith("/"):
            if len(values["imageId"]) == 37:
                return ImageType.Friend
            else:
                return ImageType.Temp
        elif values["imageId"].startswith("{") and values["imageId"].endswith(
            "}.mirai"
        ):
            return ImageType.Group
        else:
            return ImageType.Unknown

    async def toExternal(self):
        try:
            want_type = image_upload_method_type_map.get(image_method.get())
            if self.type != want_type and self.url:
                app = application.get()
                image_byte = await self.http_to_bytes()
                return await app.uploadImage(
                    image_byte, image_method.get(), return_external=True
                )
        except LookupError:
            pass
        return External.Image(imageId=self.imageId, url=self.url, path=self.path)

    @classmethod
    def fromExternal(cls, external_element) -> "Image":
        return cls(
            imageId=external_element.imageId,
            url=external_element.url,
            path=external_element.path,
        )

    @classmethod
    def fromLocalFile(
        cls, filepath: Union[Path, str], method: Optional[UploadMethods] = None
    ) -> "Image":
        """从本地文件中创建一个 Shadow Element, 以此在发送时自动上传图片至服务器, 并借此使包含的图片成功发送.

        Args:
            filepath (Union[Path, str]): 需要上传的图片路径, 可以是字符串也可以是 pathlib.Path 实例
            method (Optional[UploadMethods], default = None): 图片上传时使用的方法, 通常可以使程序自行判定.

        Raises:
            FileNotFoundError: 所描述的图片文件在文件系统中不存在.

        Returns:
            [Shadow Element]: 返回值为一合法, 但不包括任何 Image 特征属性的叠加态消息元素; 其包含有一 asFlash 方法,
                可以将当前图片转为闪照形式发送.

        Examples:
        ``` python
        await app.sendGroupMessage(group, MessageChain.create([
            Image.fromLocalFile("./image.png")
            # 发闪照则是 Image.fromLocalFile("./flashimage.png").asFlash()
            # 注意: 闪照是也只能是单独一个消息.
        ]))
        ```
        """
        if isinstance(filepath, str):
            filepath = Path(filepath)
        if not filepath.exists():
            raise FileNotFoundError("you should give us a existed file's path")
        return Image_LocalFile(filepath, method)

    @classmethod
    def fromUnsafePath(cls, path: Union[Path, str]) -> "External.Image":
        """不检查路径安全性, 直接实例化元素, 让 mirai-api-http 自行读取图片文件.

        Args:
            path (Union[Path, str]): 图片文件路径

        Returns:
            Image: 作为外部态存在的 Image 消息元素
        """
        return External.Image(path=str(path))

    @classmethod
    def fromUnsafeBytes(
        cls, image_bytes: bytes, method: Optional[UploadMethods] = None
    ) -> "Image":
        """从不保证有效性的 bytes 中创建一个 Shadow Element, 以此在发送时自动作为图片上传至服务器, 并借此使其可能包含的图片成功发送.

        Args:
            image_bytes: (bytes): 任意 bytes, 不保证内部可能的图片的有效性
            method (Optional[UploadMethods], default = None): 图片上传时使用的方法, 通常可以使程序自行判定.

        Returns:
            [Shadow Element]: 返回值为一合法, 但不包括任何 Image 特征属性的叠加态消息元素; 其包含有一 asFlash 方法,
                可以将当前图片转为闪照形式发送.
        """
        return Image_UnsafeBytes(image_bytes, method)

    @classmethod
    def fromNetworkAddress(
        cls, url: str, method: Optional[UploadMethods] = None
    ) -> "Image":
        """从不保证有效性的网络位置中创建一个 Shadow Element, 以此在发送时自动从该指定位置获取并作为图片上传至服务器,
        并借此使其可能包含的图片成功发送.

        Args:
            url: (str): 可以是任意 http/https 的 url, 不保证其有效性, 且可能抛出任意形式的网络错误
            method (Optional[UploadMethods], default = None): 图片上传时使用的方法, 通常可以使程序自行判定.

        Raises:
            ClientResponseError: HTTP 网络请求错误

        Returns:
            [Shadow Element]: 返回值为一合法, 但不包括任何 Image 特征属性的叠加态消息元素; 其包含有一 asFlash 方法,
                可以将当前图片转为闪照形式发送.
        """
        return Image_NetworkAddress(url, method)

    @classmethod
    def fromUnsafeAddress(cls, url: str) -> "External.Image":
        return External.Image(url=url)

    def asDisplay(self) -> str:
        return "[图片]"

    async def http_to_bytes(self, url: str = None) -> bytes:
        """从远端服务器获取图片的 bytes, 注意, 你无法获取并不包含 url 属性的本元素的 bytes.

        Args:
            url (str, optional): 如果提供, 则从本参数获取 bytes. 默认为 None.

        Raises:
            ValueError: 你尝试获取并不包含 url 属性的本元素的 bytes.

        Returns:
            bytes: 图片原始数据
        """
        if not (self.url or url):
            raise ValueError("you should offer a url.")
        async with ClientSession() as session:
            async with session.get(self.url or url) as response:
                response.raise_for_status()
                return await response.read()

    def asFlash(self) -> "FlashImage":
        return FlashImage.fromOriginalImage(self)

    def asSerializationString(self) -> str:
        return f"[mirai:image:{self.imageId}]"


class FlashImage(Image, InternalElement):
    """用于承载 QQ 中的特殊消息: 闪照的消息组件.

    通常, 你需要先使用 Image 中提供的各种工厂方法创建一"普通"的 Image 对象, 然后再使用 asFlash 方法将其转换为闪照.
    """

    def toExternal(self):
        return External.FlashImage(imageId=self.imageId, url=self.url, path=self.path)

    @classmethod
    def fromExternal(cls, external_element) -> "FlashImage":
        return cls(
            imageId=external_element.imageId,
            url=external_element.url,
            path=external_element.path,
        )

    @classmethod
    def fromOriginalImage(cls, image_element: Image) -> "FlashImage":
        return cls(
            imageId=image_element.imageId,
            url=image_element.url,
            path=image_element.path,
        )

    def asDisplay(self) -> str:
        return "[闪照]"

    def asNormal(self) -> "Image":
        return Image.fromExternal(self)

    def asSerializationString(self) -> str:
        return f"[mirai:flash:{self.imageId}]"


class VoiceUploadType(Enum):
    Group = "group"
    Unknown = "unknown"


class Voice_LocalFile(ShadowElement, InternalElement, ExternalElement):
    filepath: Path
    method: Optional[UploadMethods]

    def __init__(self, filepath: Path, method: Optional[UploadMethods] = None) -> None:
        super().__init__(filepath=filepath, method=method)

    def fromExternal(cls, external_element) -> "InternalElement":
        return external_element

    async def toExternal(self):
        app = application.get()
        try:
            methodd = self.method or image_method.get()
        except LookupError:
            raise ValueError(
                "you should give the 'method' for upload when you are out of the event receiver."
            )

        return await app.uploadVoice(
            self.filepath.read_bytes(), methodd, return_external=True
        )

    async def getReal(self, method: UploadMethods) -> "Image":
        """从本 Shadow Element 中生成一真正的 Voice 对象.
        Args:
            method (UploadMethods): 所需求的图片的上传类型, 具体请阅读 UploadMethods 的相关文档.
        Raises:
            ClientResponseError: HTTP 网络请求错误
        Returns:
            Voice: 所生成的, 真正的 Voice 对象.
        """
        app = application.get()
        return await app.uploadVoice(
            self.filepath.read_bytes(), method, return_external=True
        )


class Voice(InternalElement):
    voiceId: Optional[str] = None
    url: Optional[str] = None
    path: Optional[str] = None
    type: Optional[VoiceUploadType]

    def toExternal(self):
        return External.Voice(voiceId=self.voiceId, url=self.url, path=self.path)

    @classmethod
    def fromExternal(cls, external_element) -> "Voice":
        return cls(
            voiceId=external_element.voiceId,
            url=external_element.url,
            path=external_element.path,
        )

    def asSerializationString(self) -> str:
        return f"[mirai:voice:{self.voiceId}]"

    def asDisplay(self) -> str:
        return "[语音]"

    @validator("type", always=True, allow_reuse=True)
    def _(cls, v, values) -> VoiceUploadType:
        if v:
            return v
        if "voiceId" not in values:
            return VoiceUploadType.Unknown
        if not values["voiceId"]:
            return VoiceUploadType.Unknown
        if values["voiceId"]:
            return VoiceUploadType.Group  # mirai 当前版本只支持群语音.

    def fromLocalFile(
        self, filepath: Union[str, Path], method: Optional[UploadMethods] = None
    ) -> "Voice":
        """从本地文件中创建一个 Shadow Element, 以此在发送时自动上传语音至服务器, 并借此使包含的语音成功发送.

        Args:
            filepath (Union[Path, str]): 需要上传的图片路径, 可以是字符串也可以是 pathlib.Path 实例
            method (Optional[UploadMethods], default = None): 语音上传时使用的方法, 通常可以使程序自行判定.

        Raises:
            FileNotFoundError: 所描述的语音文件在文件系统中不存在.

        Returns:
            [Shadow Element]: 返回值为一合法, 但不包括任何 Voice 特征属性的叠加态消息元素
        """
        if isinstance(filepath, str):
            filepath = Path(filepath)
        if not filepath.exists():
            raise FileNotFoundError("you should give us a existed file's path")
        return Voice_LocalFile(filepath, method)


class Xml(InternalElement, ExternalElement):
    type = "Xml"
    xml: str

    def __init__(self, xml, *_, **__) -> None:
        super().__init__(xml=xml)

    def toExternal(self):
        return self

    @classmethod
    def fromExternal(cls, external_element) -> "Xml":
        return external_element

    def asDisplay(self) -> str:
        return "[XML消息]"


class Json(InternalElement, ExternalElement):
    type = "Json"
    Json: str = Field(..., alias="json")

    def __init__(self, json: Union[dict, str], *_, **__) -> None:
        if isinstance(json, dict):
            json = MJson.dumps(json)
        super().__init__(json=json)

    def dict(self, *args, **kwargs):
        return super().dict(*args, **({**kwargs, "by_alias": True}))

    def asDisplay(self) -> str:
        return "[JSON消息]"

    def toExternal(self) -> "Json":
        return self

    @classmethod
    def fromExternal(_, external_element) -> "Json":
        return external_element


class App(InternalElement, ExternalElement):
    type = "App"
    content: str

    def asDisplay(self) -> str:
        return "[APP消息]"

    def toExternal(self) -> "App":
        return self

    @classmethod
    def fromExternal(_, external_element) -> "App":
        return external_element


class PokeMethods(Enum):
    Poke = "Poke"
    ShowLove = "ShowLove"
    Like = "Like"
    Heartbroken = "Heartbroken"
    SixSixSix = "SixSixSix"
    FangDaZhao = "FangDaZhao"


class Poke(InternalElement, ExternalElement):
    type = "Poke"
    name: PokeMethods

    def asDisplay(self) -> str:
        return "[戳一戳:{0}]".format(self.name)

    def toExternal(self) -> "Poke":
        return self

    @classmethod
    def fromExternal(_, external_element) -> "Poke":
        return external_element
'''



import os
import json
from functools import partial
from typing import Optional

import requests
from urllib.parse import urljoin
from urllib.parse import quote
from tenacity import retry

from .httpapi import MiraiHttpApi
from .session import MiraiSession
from .message.chain import MessageChain
from .message.messages import Message


class MiraiApi(object):
    def __init__(self, host: str, authKey: str):
        self.host = host
        self.authKey = authKey
        self.llapi = MiraiHttpApi(host)

    def get_session(self, account: str):
        return MiraiSession(self.llapi, account, self.authKey)
